@@ -31,12 +31,14 @@ import time
 import gflags
 
 import git
+import git_owners
 
 # Used mainly to provide a usage summary with -h, consistent with
 # other git commands.
 gflags.DEFINE_bool('help-summary', False, 'Show a short usage message and exit.', short_name='h')
 
 gflags.DEFINE_list('reviewers', list(), 'Comma-separated list of reviewers.', short_name='r')
+gflags.DEFINE_bool('ignore-owners', False, 'Set to ignore OWNERS files for this particular change.')
 gflags.DEFINE_list('cc', list(),
                    'Comma-separated list of addresses to copy on change notification mails.')
 gflags.DEFINE_string('bug', None, 'Bug ID to include in the commit message header', short_name='b')
@@ -71,6 +73,18 @@ gflags.DEFINE_bool('fake-push', False,
 
 FLAGS = gflags.FLAGS
 
+BRANCH_SHORT_LENGTH = 7
+
+BASH_COLORS = {
+    'YELLOW': '\033[93m',
+    'GRAY': '\033[30m',
+    'CLEAR': '\033[0m'
+}
+COLOR_CURRENT = BASH_COLORS['YELLOW']  # Current git-change branch.
+COLOR_OBSOLETE = BASH_COLORS['GRAY']  # Change branch that has been merged to master.
+COLOR_CLEAR = BASH_COLORS['CLEAR']  # Reset color to terminal default.
+
+
 
 def usage(include_flags=True):
     """Prints a usage message.
@@ -85,11 +99,12 @@ def usage(include_flags=True):
                '   or: git change submit\n'
                '   or: git change gc\n'
                '\n'
-               '<create-options>: [-r|--reviewers=] [--cc=] [-b|--bug=] [-m|--message=] '
-               '[--topic=] [--fetch] [--switch] [--chain] '
+               '<create-options>: [-r|--reviewers=] [--ignore-owners=] [--cc=] [-b|--bug=] '
+               '[-m|--message=] [--topic=] [--fetch] [--switch] [--chain] '
                '[--use-head-commit] [--merge-commit] [--skip=]\n'
                '\n'
-               '<update-options>[-r|--reviewers=] [--cc=] [-b|--bug=] [--skip=]\n'
+               '<update-options>[-r|--reviewers=] [--ignore-owners=] [--cc=] '
+               ' [-b|--bug=] [--skip=]\n'
                '\n'
                'See git-change(1) for full documentation.')
     print message
@@ -104,7 +119,7 @@ def exit_error(message, prefix='Error: ', status=1):
         message: A string representing the error message to
             print. 'Error: ' will be prepended to message.
         prefix: A string to prepend to message
-        stautus: An integer representing the exit status code.
+        status: An integer representing the exit status code.
     """
     sys.stderr.write('%s%s\n' % (prefix, message))
     sys.exit(status)
@@ -122,25 +137,29 @@ def get_change_id_from_branch():
         branch name, or None if the current branch is not a temporary
         change branch.
     """
-    branch = git.get_branch()
+    branch = git.get_current_branch()
     if branch.startswith('change-I'):
         _, change_id = branch.split('-')
         return change_id
     return None
 
 
-def get_change_id_from_head():
-    """Returns the change ID of the last commit.
+def get_change_id_from_commit(commit):
+    """Returns the change ID of the given commit.
 
-    If the change ID is available as the value of the Change-Id
-    attribute in the HEAD commit message, that value is returned. If
-    there is no Change-Id attribute, returns None.
+    If the change ID is available as the value of the Change-Id header
+    in the given commit's message, that value is returned. If there is
+    no Change-Id header, returns None.
+
+    Args:
+        commit: A string representing the commit from which to read
+            the change ID.
 
     Returns:
-        A string representing the HEAD commit's change ID if it is
+        A string representing the given commit's change ID if it is
         available, or None if not.
     """
-    output = git.run_command('git cat-file -p HEAD', trap_stdout=True)
+    output = git.run_command('git cat-file -p %s' % commit, trap_stdout=True)
     lines = output.split('\n')
     for line in lines:
         if line.startswith('Change-Id:'):
@@ -149,6 +168,41 @@ def get_change_id_from_head():
     return None
 
 
+def get_change_id_from_head():
+    """Returns the change ID of the HEAD commit.
+
+    Returns:
+        A string representing the HEAD commit's change ID if it is
+        available, or None if not.
+    """
+    return get_change_id_from_commit('HEAD')
+
+
+def get_reviewers_for_change():
+    """Gets the reviewers for this change from command flag and OWNERS files.
+
+    Combines two sets of Gerrit reviewer usernames to create one set of
+    reviewers for this change:
+        1. Reviewer usernames passed in with the command line flag 'reviewers'.
+        2. Reviewer usernames listed in relevant OWNERS files, if the repo is
+            configured for OWNERS files and the commit author has not passed the
+            flag 'ignore-owners=True'. See the git_owners module for more
+            information about OWNERS files.
+
+    Returns:
+        A list of strings representing Gerrit Code Review usernames.
+    """
+    reviewers = set()
+    reviewers.update(FLAGS.reviewers)
+
+    repo_configured_for_owners = git.get_config_option('git-change.include-owners') == 'true'
+    ignore_owners_flag = FLAGS['ignore-owners'].value
+    if repo_configured_for_owners and not ignore_owners_flag:
+        reviewers.update(git_owners.get_change_owners())
+
+    return [r for r in reviewers if r]
+
+
 def build_push_command(branch):
     """Builds a git push command string for pushing a Gerrit change.
 
@@ -163,9 +217,9 @@ def build_push_command(branch):
     """
     command = 'git push %s' % FLAGS.remote
     receive_pack_args = []
-    for reviewer in FLAGS.reviewers:
-        if reviewer:  # trailing commas in flag value generate blank entries
-            receive_pack_args.append('--reviewer=%s' % reviewer)
+
+    for reviewer in get_reviewers_for_change():
+        receive_pack_args.append('--reviewer=%s' % reviewer)
     for cc in FLAGS.cc:
         if cc:  # trailing commas in flag value generate blank entries
             receive_pack_args.append('--cc=%s' % cc)
@@ -197,16 +251,23 @@ def check_unmerged_commits(branch):
         remote branch. False if not, or if the user elected to proceed
         anyway.
     """
-    output = git.run_command('git log --oneline %s ^%s/%s --' % (branch, FLAGS.remote, branch),
-                             trap_stdout=True)
+    # Gather output of git log.
+    output = git.run_command('git log --oneline %s ^%s/%s --' % (
+        branch, FLAGS.remote, branch), trap_stdout=True).strip()
+
+    # Count how many commits ahead we are vs. origin:
+    commits_ahead = int(git.run_command('git rev-list %s ^%s/%s --count' % (
+        branch, FLAGS.remote, branch), trap_stdout=True).strip())
+
     if not output or FLAGS['dry-run'].value:
         return False
 
-    print 'Your branch %s is ahead of its remote by the following %s commit(s):\n' % (
-        branch, len(output.split('\n')))
+    print 'Your branch %s is ahead of its remote by the following %i commit%s:\n' % (
+        branch, commits_ahead, 's' if commits_ahead > 1 else '')
+
     sys.stdout.write(output)
     user_input = raw_input(
-        '\nIf we continue, each of the commits above may result in a new code\n'
+        '\n\nIf we continue, each of the commits above may result in a new code\n'
         'review and a submit dependency in Gerrit. You might try syncing the\n'
         'remote branch by passing the --fetch flag.\n'
         'Continue? ')
@@ -224,7 +285,7 @@ def get_change(change_id):
     created from the JSON search result.
 
     This function exits with a non-zero status if the Gerrit search
-    returns multiple results for change_id.
+    returns zero or multiple results for change_id.
 
     Args:
         change_id: A string representing the ID of the desired change.
@@ -274,6 +335,9 @@ def update_change():
     ID is taken from the current branch, which should be a temporary
     change branch created by a previous run of git-change.
     """
+    if FLAGS.message is not None:
+        exit_error('--message cannot be used with the update subcommand.')
+
     change_id = check_for_change_branch()
     change = get_change(change_id)
     if not change['open']:
@@ -285,7 +349,6 @@ def update_change():
     # we have a new patch set.
     if (FLAGS.reviewers or FLAGS.cc or FLAGS.bug is not None or
         git.run_command('git diff --cached --name-status', trap_stdout=True)):
-
         commit_change(['--amend'])
 
     command = build_push_command(change['branch'])
@@ -362,6 +425,33 @@ def sanity_check_merge_commit():
             sys.exit(1)
 
 
+def get_target_branch(commit='HEAD'):
+    """Returns the Git-change target branch of the given commit.
+
+    The target branch is read from the given commit's note. If the
+    target branch is not available from the note, an attempt is made
+    to read the Change-Id header from the given commit's message,
+    which is then used to fetch the target branch from Gerrit.
+
+    Args:
+        commit: A string representing the commit whose target branch
+            is desired.
+
+    Returns:
+        A string representing the name of the target branch if
+        available, otherwise None.
+    """
+    note_data = git.read_note(commit)
+    target_branch = note_data.get('Target-Branch', None)
+    if target_branch is None:
+        change_id = get_change_id_from_commit(commit)
+        if change_id is None:
+            return None
+        change = get_change(change_id)
+        target_branch = change['branch']
+    return target_branch
+
+
 def determine_branches():
     """Determines the current and target branches.
 
@@ -378,29 +468,30 @@ def determine_branches():
         A tuple of two strings: the name of the current branch and
         that of the target branch.
     """
-    original_branch = git.get_branch()
+    current_branch = git.get_current_branch()
     if FLAGS.chain:
-        # Extract the target branch from the previous change with
-        # which we're chaining.
-        previous_change_id = get_change_id_from_branch()
-        if previous_change_id is None:
+        # Extract the target branch from the current change branch.
+        target_branch = get_target_branch()
+        if target_branch is None:
             exit_error('The current branch must be a change branch when you specify --chain.')
-        previous_change = get_change(previous_change_id)
-        target_branch = previous_change['branch']
     else:
-        if original_branch.startswith('change-I'):
+        if current_branch.startswith('change-I'):
             exit_error('You are in a temporary change branch. '
                        'If you wish to chain commits, pass --chain.')
-        target_branch = original_branch
+        target_branch = current_branch
 
-    return original_branch, target_branch
+    return current_branch, target_branch
 
 
 def commit_staged_changes(original_branch, tmp_branch):
     """Commits staged changes.
 
     A change ID will be generated by the commit-msg hook as a
-    side-effect. This function exits the process on errors.
+    side-effect.
+
+    If the git-commit command fails or is interrupted by the user
+    (e.g., with Control-C) the original branch is restored and the
+    process exits with a non-zero status.
 
     Args:
         original_branch: A string representing the name of the branch
@@ -469,13 +560,24 @@ def create_change():
         print ('\nWARNING: Reading change ID from the HEAD commit failed. (You may need to\n'
                'install the Gerrit commit-msg hook.) Before continuing, you need to add\n'
                'the change ID header to the HEAD commit message (git commit --amend) and\n'
-               'rename the branch %s to change-<change-ID> manaully.' % tmp_branch)
+               'rename the branch %s to change-<change-ID> manually.' % tmp_branch)
         new_branch = tmp_branch
     else:
         new_branch = 'change-%s' % change_id
         git.run_command('git branch -m %s %s' % (tmp_branch, new_branch))
     print '\nCreated branch: %s\n' % new_branch
 
+    # Cache change meta-data in a note. With --chain, Parent-Branch is
+    # the temporary change branch that is the base of the
+    # chain. Without --chain, Parent-Branch and Taget-Branch are the
+    # same.
+    note = {
+        'Change-Id': change_id,
+        'Target-Branch': target_branch,
+        'Parent-Branch': original_branch,
+        }
+    git.write_note(note)
+
     command = build_push_command(target_branch)
     try:
         git.run_command(command)
@@ -521,10 +623,9 @@ def rebase():
     If there are conflicts with either rebase operation, the process
     terminates and it is up to the user to resolve the conflicts.
     """
-    change_id = check_for_change_branch()
-    change = get_change(change_id)
-    target_branch = change['branch']
-    change_branch = git.get_branch()
+    check_for_change_branch()
+    target_branch = get_target_branch()
+    change_branch = git.get_current_branch()
 
     git.run_command_or_die('git checkout %s' % target_branch)
     try:
@@ -575,12 +676,37 @@ def list_change_branches():
     if not branches:
         print 'You have no change branches to list'
         return
+
+    not_merged_branches = git.run_command('git branch --no-merged',
+                                          trap_stdout=True).strip().split('\n')
+    not_merged_branches = [line.strip()[BRANCH_SHORT_LENGTH:] for line in not_merged_branches]
+
     print 'Change branches:\n'
     i = 0
     for branch in branches:
         i += 1
+
         output = git.run_command('git log --oneline -1 %s --' % branch, trap_stdout=True)
-        sys.stdout.write('{0:>2}. {1} {2}'.format(i, branch, output))
+        change_id = output.split(' ')[0]
+        description = ' '.join(output.split(' ')[1:])
+        short_branch = branch[0:16]
+        change_branch = branch.split('-')[1]
+
+        # handle colors here
+        use_color = git.get_config_option('git-change.color')
+        use_color = (use_color != 'false')  # auto or yes or anything else count as True
+
+        cid_url = git.get_config_option('git-change.cid-url') or ''
+
+        if use_color and change_branch not in not_merged_branches:  # not not == is merged
+            sys.stdout.write(COLOR_OBSOLETE)
+
+        if use_color and change_branch == get_change_id_from_branch():
+            sys.stdout.write(COLOR_CURRENT)
+        sys.stdout.write('{i:>2}: {branch_id} {href}{cid} {name}'.format(
+            i=i, branch_id=short_branch, href=cid_url, cid=change_id, name=description))
+        if use_color:
+            sys.stdout.write(COLOR_CLEAR)
     try:
         selection = raw_input('\nSelect a branch number to check out, '
                               'or hit enter to exit: ')
@@ -638,14 +764,27 @@ def print_push_command():
         change = get_change(change_id)
         target_branch = change['branch']
     else:
-        target_branch = git.get_branch()
+        target_branch = git.get_current_branch()
     print build_push_command(target_branch)
 
 
-def main(argv):
-    if FLAGS['help-summary'].value:
-        usage(include_flags=False)
-        sys.exit()
+def configure():
+    """Configures git-change.
+
+    Verifies that required configuration options are in place and
+    merges command-line flags with config options.
+    """
+    # Make sure the notes.rewriteRef config option contains the
+    # git-change notes ref so that external commands that rewrite
+    # commits (e.g. git-commit --amend) copy notes to the rewritten
+    # commits.
+    rewrite_ref = git.get_config_option('notes.rewriteRef')
+    if not rewrite_ref:
+        rewrite_ref = git.NOTES_REF
+        git.set_config_option('notes.rewriteRef', rewrite_ref)
+    elif not git.NOTES_REF in rewrite_ref:
+        rewrite_ref = '%s:%s' % (rewrite_ref, git.NOTES_REF)
+        git.set_config_option('notes.rewriteRef', rewrite_ref)
 
     # Get remote from command-line flag or config option, otherwise
     # fall back to flag default.
@@ -667,9 +806,17 @@ def main(argv):
     if FLAGS['merge-commit'].value:
         FLAGS['use-head-commit'].value = True
 
+
+def main(argv):
+    if FLAGS['help-summary'].value:
+        usage(include_flags=False)
+        sys.exit()
+
     # Fail gracefully if run outside a git repository.
     git.run_command_or_die('git status')
 
+    configure()
+
     argc = len(argv)
     if argc > 2:
         usage(include_flags=False)
